#!/usr/bin/env python3
# -*- coding: utf-8 -*-

"""Execution Context Module."""

import logging
import math
import re
from typing import TYPE_CHECKING, Any, Dict, Optional, Union

from biolib.pyppeteer.pyppeteer import helpers
from biolib.pyppeteer.pyppeteer.connection import CDPSession
from biolib.pyppeteer.pyppeteer.errors import ElementHandleError
from biolib.pyppeteer.pyppeteer.jshandle import ElementHandle, JSHandle, createJSHandle
from biolib.pyppeteer.pyppeteer.models import JSFunctionArg

if TYPE_CHECKING:
    from pyppeteer.frame import Frame
    from pyppeteer.domworld import DOMWorld

logger = logging.getLogger(__name__)

EVALUATION_SCRIPT_URL = '__pyppeteer_evaluation_script__'
SOURCE_URL_REGEX = re.compile(r'^[ \t]*//[@#] sourceURL=\s*(\S*?)\s*$', re.MULTILINE,)

class ExecutionContext:
    """Execution Context class."""

    def __init__(self, client: CDPSession, contextPayload: Dict, world: Optional['DOMWorld']) -> None:
        self._client = client
        self._world = world
        self._contextId = contextPayload['id']

    @property
    def frame(self) -> Optional['Frame']:
        """Return frame associated with this execution context."""
        if self._world:
            return self._world.frame

    async def evaluate(self, pageFunction: str, *args: JSFunctionArg) -> JSFunctionArg:
        """Execute ``pageFunction`` on this context.

        Details see :meth:`pyppeteer.page.Page.evaluate`.
        """
        return await self._evaluateInternal(True, pageFunction, *args)

    async def evaluateHandle(self, pageFunction: str, *args: JSFunctionArg) -> Union['JSHandle', 'ElementHandle']:
        """Execute ``pageFunction`` on this context.
        Details see :meth:`pyppeteer.page.Page.evaluateHandle`.
        """
        return await self._evaluateInternal(False, pageFunction, *args)

    async def _evaluateInternal(
        self, returnByValue: bool, pageFunction: str, *args: JSFunctionArg
    ) -> Union['JSHandle', 'ElementHandle', JSFunctionArg]:
        suffix = f'//# sourceURL={EVALUATION_SCRIPT_URL}'

        if not helpers.is_js_func(pageFunction):
            try:
                if SOURCE_URL_REGEX.match(pageFunction):
                    expressionWithSourceUrl = pageFunction
                else:
                    expressionWithSourceUrl = f'{pageFunction}\n{suffix}'
                remoteObject = await self._client.send(
                    'Runtime.evaluate',
                    {
                        'expression': expressionWithSourceUrl,
                        'contextId': self._contextId,
                        'returnByValue': returnByValue,
                        'awaitPromise': True,
                        'userGesture': True,
                    },
                )
            except Exception as e:
                exceptionDetails = rewriteError(e)
                raise type(e)(f'Evaluation failed: {helpers.getExceptionMessage(exceptionDetails)}')
        else:
            try:
                remoteObject = await self._client.send(
                    'Runtime.callFunctionOn',
                    {
                        'functionDeclaration': f'{pageFunction}\n{suffix}\n',
                        'executionContextId': self._contextId,
                        'arguments': [self._convertArgument(arg) for arg in args],
                        'returnByValue': returnByValue,
                        'awaitPromise': True,
                        'userGesture': True,
                    },
                )
            except Exception as e:
                exceptionDetails = rewriteError(e)
                raise type(e)(f'Evaluation failed: {helpers.getExceptionMessage(exceptionDetails)}')

        exceptionDetails = remoteObject.get('exceptionDetails')
        if exceptionDetails:
            raise ElementHandleError('Evaluation failed: {}'.format(helpers.getExceptionMessage(exceptionDetails)))

        remoteObject = remoteObject['result']
        if returnByValue:
            return helpers.valueFromRemoteObject(remoteObject)
        else:
            return createJSHandle(self, remoteObject)

    def _convertArgument(self, arg: Any) -> Dict:  # noqa: C901
        if arg == math.inf:
            return {'unserializableValue': 'Infinity'}
        if arg == -math.inf:
            return {'unserializableValue': '-Infinity'}
        objectHandle = arg if isinstance(arg, JSHandle) else None
        if objectHandle:
            if objectHandle._context != self:
                raise ElementHandleError('JSHandles can be evaluated only in the context they were created!')
            if objectHandle._disposed:
                raise ElementHandleError('JSHandle is disposed!')
            if objectHandle._remoteObject.get('unserializableValue'):
                return {'unserializableValue': objectHandle._remoteObject.get('unserializableValue')}
            if not objectHandle._remoteObject.get('objectId'):
                return {'value': objectHandle._remoteObject.get('value')}
            return {'objectId': objectHandle._remoteObject.get('objectId')}
        return {'value': arg}

    async def queryObjects(self, prototypeHandle: 'JSHandle'):
        if prototypeHandle._disposed:
            raise ElementHandleError('Prototype JSHandle is disposed')
        if not prototypeHandle._remoteObject.get('objectId'):
            raise ElementHandleError('Prototype JSHandle must not be referencing primitive value')
        response = await self._client.send(
            'Runtime.queryObjects', {'prototypeObjectId': prototypeHandle._remoteObject['objectId']}
        )
        return createJSHandle(context=self, remoteObject=response.get('objects'))

    async def _adoptBackendNodeId(self, backendNodeId: int):
        obj = await self._client.send(
            'DOM.resolveNode', {'backendNodeId': backendNodeId, 'executionContextId': self._contextId}
        )
        obj = obj['object']
        return createJSHandle(context=self, remoteObject=obj)

    async def _adoptElementHandle(self, elementHandle: ElementHandle):
        if elementHandle.executionContext == self:
            raise ElementHandleError('Cannot adopt handle that already belongs to this execution context')
        if not self._world:
            raise ElementHandleError('Cannot adopt handle without DOMWorld')
        nodeInfo = await self._client.send('DOM.describeNode', {'objectId': elementHandle._remoteObject['objectId']})
        return await self._adoptBackendNodeId(nodeInfo['node']['backendNodeId'])


def rewriteError(error: Exception) -> Union[None, Dict[str, Dict[str, str]]]:
    msg = error.args[0]
    if 'Object reference chain is too long' in msg:
        return {'result': {'type': 'undefined'}}
    if "Object couldn't be returned by value" in msg:
        return {'result': {'type': 'undefined'}}
    if msg.endswith('Cannot find context with specified id'):
        raise type(error)('Execution context was destroyed, most likely because of a navigation.')
    raise error



import asyncio
from biolib.pyppeteer.pyppeteer import launch, command
from biolib.pyppeteer.pyppeteer.launcher import resolveExecutablePath, __chromium_revision__
import http.server
import socketserver
import threading
import base64
import os
import sys
import subprocess
import re
import requests
import logging
import shutil
from pathlib import Path

# fix required for async to work in notebooks
import nest_asyncio
nest_asyncio.apply()

import sys, logging
logging.basicConfig(format='%(asctime)s | %(levelname)s : %(message)s',
                     level=logging.INFO, stream=sys.stdout)
logging.getLogger('biolib.pyppeteer.pyppeteer').setLevel(logging.ERROR)
logging.getLogger('biolib.pyppeteer.pyppeteer.connection').setLevel(logging.ERROR)

biolib_logger = logging.getLogger('biolib')

class BioLibError(Exception):
    def __init__(self, message):
        self.message = message


class CompletedProcess:
    def __init__(self, stdout, stderr, exitcode):
        self.stdout = stdout
        self.stderr = stderr
        self.exitcode = exitcode

    def ipython_markdown(self):
        from IPython.display import display, Markdown
        markdown_str = self.stdout.decode("utf-8")
        # prepend ./biolib_results/ to all paths
        # ie [SeqLogo](./SeqLogo2.png) test ![SeqLogo](./SeqLogo.png)
        # ![SeqLogo](SeqLogo.png)  ![SeqLogo](/SeqLogo.png)
        # is transformed to ![SeqLogo](./biolib_results/SeqLogo2.png) test ![SeqLogo](./biolib_results/SeqLogo.png)
        # ![SeqLogo](./biolib_results/SeqLogo.png)  ![SeqLogo](./biolib_results/SeqLogo.png)
        markdown_str_modified = re.sub('\!\[([^\]]*)\]\((\.\/|\/|)([^\)]*)\)', '![\\1](./biolib_results/\\3)', markdown_str)
        display(Markdown(markdown_str_modified))

class BioLibServer(socketserver.TCPServer):
    log_level = logging.INFO

class BioLibHandler(http.server.SimpleHTTPRequestHandler):
    def log_message(self, format, *args):
        if self.server.log_level == BioLib.TRACE_LOGGING:
            http.server.SimpleHTTPRequestHandler.log_message(self, format, *args)


def js_to_python_byte_array_converter(js_encoded):
    return bytes(list([js_encoded[str(i)] for i in range(len(js_encoded))]))


def python_bytes_to_byte64(data):
    return base64.b64encode(data).decode('ascii')


class BioLib:
    executable_path = None
    no_sandbox = True
    log_level = logging.INFO
    TRACE_LOGGING = 50
    auth_email = None
    auth_password = None
    auth_access_token = None
    auth_refresh_token = None
    host = "https://biolib.com"

    @staticmethod
    def set_chrome_path(path):
        BioLib.executable_path = path

    @staticmethod
    def set_sandbox(use_sandbox):
        BioLib.no_sandbox = not use_sandbox

    @staticmethod
    def set_credentials(email, password):
        BioLib.auth_email = email
        BioLib.auth_password = password

    @staticmethod
    def attempt_login(exit_on_failure=False):
        if BioLib.auth_email is None:
            if exit_on_failure:
                biolib_logger.error("Error: Attempted login, but BIOLIB_EMAIL was not set, exiting...")
                sys.exit(1)
            else:
                biolib_logger.debug("Attempted login, but BIOLIB_EMAIL was not set, so continuing without logging in")
                return
        if BioLib.auth_password is None:
            if exit_on_failure:
                biolib_logger.error("Error: Attempted login, but BIOLIB_PASSWORD was not set, exiting...")
                sys.exit(1)
            else:
                biolib_logger.debug("Attempted login, but BIOLIB_PASSWORD was not set, so continuing without logging in")
                return

        r = requests.post(f"{BioLib.host}/api/user/token/",
                          json={"email": BioLib.auth_email, "password": BioLib.auth_password })
        response = r.json()
        if not r.ok:
            biolib_logger.error(f"Login failed for {BioLib.auth_email}:")
            biolib_logger.error(response["detail"])
            sys.exit(1)
        else:
            BioLib.auth_refresh_token = response["refresh"]
            BioLib.auth_access_token = response["access"]
            biolib_logger.info(f"Successfully authenticated {BioLib.auth_email}")

    @staticmethod
    def push(author_and_app_name, app_path):
        # prepare zip file
        if not Path(f"{app_path}/.biolib/config.yml").is_file():
            biolib_logger.error("The file .biolib/config.yml was not found in the application directory")
            sys.exit(1)
        cwd = os.getcwd()
        os.chdir(app_path)
        zip_name = "biolib-cli-build-tmp-zip-1js7ah2akd"
        zip_filename = f"{zip_name}.zip"
        if Path(zip_filename).is_file():
            biolib_logger.error(f"Please make sure you don't have a file "
                                f"called {zip_filename} in the application directory")
            sys.exit(1)
        shutil.make_archive(zip_name, 'zip', root_dir="..", base_dir=app_path)
        f = open(zip_filename, "rb")
        zip_binary = f.read()
        f.close()
        os.remove(zip_filename)
        # change back to old working directory
        os.chdir(cwd)

        # login and get app data
        BioLib.attempt_login(exit_on_failure=True)
        author_and_app_name_split = author_and_app_name.split("/")
        app = BioLib.get_app_from_author_and_app_name(author_and_app_name_split[0],
                                                      author_and_app_name_split[1])
        # push new app version
        r = requests.post(
            f"{BioLib.host}/api/app_versions/", 
            headers={"Authorization": f"Bearer {BioLib.auth_access_token}"},
            files={
                "app": (None, app["public_id"]),
                "set_as_active": (None, "true"),
                "state": (None, "published"),
                "source_files_zip": zip_binary
            })
        if not r.ok:
            biolib_logger.error(f"Push failed for {author_and_app_name}:")
            biolib_logger.error(response)
            sys.exit(1)
        else:
            # TODO: When response includes the version number, print the URL for the new app version
            _response = r.json()
            biolib_logger.info(f"Successfully pushed app version for {author_and_app_name}.")

    @staticmethod
    def get_app_from_author_and_app_name(author_name, app_name):
        headers = {}
        if BioLib.auth_access_token is not None:
            headers["Authorization"] = f"Bearer {BioLib.auth_access_token}"
        r = requests.get(f"{BioLib.host}/api/apps/?account_handle={author_name}&app_name={app_name}",
                         headers=headers)
        results = r.json()['results']
        if len(results) == 0:
            raise Exception(f"Application '{author_name}/{app_name}' was not found. "
                            f"Make sure you have spelled the application name correctly, "
                            f"and that you have the required permissions.")
        return results[0]

    @staticmethod
    def set_logging(log_level):
        BioLib.log_level = log_level
        # only activate debug logging globally if user selected trace logging
        if log_level == BioLib.TRACE_LOGGING:
            logging.getLogger().setLevel(logging.DEBUG)
        elif log_level == logging.DEBUG:
            logging.getLogger().setLevel(logging.INFO)
            biolib_logger.setLevel(logging.DEBUG)
        else:
            logging.getLogger().setLevel(log_level)

# set default logging to info
BioLib.set_logging(logging.INFO)

class BioLibApp:

    app_author = None
    app_name = None

    def __init__(self, author, name):
        self.app_author = author
        self.app_name = name

    async def call_pyppeteer(self, port, args, stdin, files, output_path):
        if not BioLib.executable_path:
            mac_chrome_path = "/Applications/Google Chrome.app/Contents/MacOS/Google Chrome"
            if os.path.isfile(mac_chrome_path):
                BioLib.set_chrome_path(mac_chrome_path)

        if not BioLib.executable_path:
            linux_chrome_path = "/usr/lib/chromium-browser/chromium-browser"

            # special install for google colab
            if not os.path.isfile(linux_chrome_path) and 'google.colab' in sys.modules:
                subprocess.run("apt-get update", shell=True, check=True)
                subprocess.run("apt install chromium-chromedriver", shell=True, check=True)

            if os.path.isfile(linux_chrome_path):
                BioLib.set_chrome_path(linux_chrome_path)

        resolved_path = resolveExecutablePath(None, __chromium_revision__)
        if not BioLib.executable_path and resolved_path[1]:
            # if executable_path is not set explicit,
            # and resolveExecutablePath failed (== we got an error message back in resolved_path[1])
            logging.info("Installing dependencies...")
            os.environ['PYPPETEER_NO_PROGRESS_BAR'] = "true"
            command.install()

        logging.info("Computing...")

        chrome_arguments = [
                '--disable-web-security',
            ]
        if BioLib.no_sandbox:
            chrome_arguments.append('--no-sandbox')

        browser = await launch(args=chrome_arguments, executablePath=BioLib.executable_path)

        # start new page
        page = await browser.newPage()

        await page.goto("http://localhost:" + str(port))

        def getCode():
            return python_bytes_to_byte64(b"code")

        def getData():
            data = {}
            data["stdin"] = stdin
            data["parameters"] = args
            data["files"] = files
            return data

        def setProgressCompute(x):
            biolib_logger.debug(f"Compute progress: {x}")

        def setProgressInitialization(x):
            biolib_logger.debug(f"Initialization progress: {x}")

        def addLogMessage(x):
            biolib_logger.debug(f"Log message: {x}")

        await page.exposeFunction("getCode", getCode)
        await page.exposeFunction("getData", getData)
        await page.exposeFunction("setProgressCompute", setProgressCompute)
        await page.exposeFunction("setProgressInitialization", setProgressInitialization)
        await page.exposeFunction("addLogMessage", addLogMessage)

        refreshToken = "undefined"
        if BioLib.auth_refresh_token:
            refreshToken = f"\"{BioLib.auth_refresh_token}\""

        result = await page.evaluate("""
        async function() {
          const refreshToken = """+refreshToken+""";
          window.BioLib.BioLibSingleton.setConfig({ baseUrl: \"""" + BioLib.host + """\", refreshToken });
          window.BioLib.AppClient.setApiClient(window.BioLib.BioLibSingleton.get());
          var data = await window.getData()
          data["stdin"] = Uint8Array.from(atob(data["stdin"]), c => c.charCodeAt(0))
          for(var i = 0; i<data["files"].length; i++) {
              data["files"][i].data = Uint8Array.from(atob(data["files"][i].data), c => c.charCodeAt(0))
          }
          var jobUtils = {}
          jobUtils["setProgressCompute"] = window.setProgressCompute
          jobUtils["setProgressInitialization"] = window.setProgressInitialization
          jobUtils["addLogMessage"] = window.addLogMessage

          try {
            return await window.BioLib.AppClient.runApp(""" + f"\"{self.app_version_uuid}\",\"{self.semantic_version}\"" + """, data, jobUtils)
          } catch(err) {
            return err.toString()
          }
        }
        """)
        logging.debug("Closing browser")
        await browser.close()

        if isinstance(result, dict):
            if len(result["files"]) > 0:
                logging.info(f"Writing output files to: {output_path}")
                os.makedirs(output_path, exist_ok=True)
                for f in result["files"]:
                    filename = output_path + f["pathWithoutFilename"] + f["name"]
                    data = js_to_python_byte_array_converter(f["data"])
                    fp = open(filename, "wb")
                    fp.write(data)
                    fp.close()
            return CompletedProcess(
                stdout=js_to_python_byte_array_converter(result["stdout"]),
                stderr=js_to_python_byte_array_converter(result["stderr"]),
                exitcode=result["exitcode"]
            )
        else:
            raise BioLibError(result)

    def __call__(self, args=[], stdin=None, files=None, output_path="biolib_results"):
        BioLib.attempt_login()
        logging.info("Loading package...")
        app = BioLib.get_app_from_author_and_app_name(self.app_author, self.app_name)
        self.app_version_uuid = app["public_id"]
        self.semantic_version = f"{app['active_version']['major']}.{app['active_version']['minor']}.{app['active_version']['patch']}"
        logging.info(f"Loaded package: {self.app_author}/{self.app_name}")

        cwd = os.getcwd()

        if stdin is None:
            stdin = b""

        if not output_path.startswith("/"):
            # output_path is relative, make absolute
            output_path = f"{cwd}/{output_path}"

        if isinstance(args, str):
            args = list(filter(lambda p: p != "", args.split(" ")))

        if not isinstance(args, list):
            raise Exception("The given input arguments must be list or str")

        if isinstance(stdin, str):
            stdin = stdin.encode("utf-8")

        if files is None:
            files = []
            for idx, arg in enumerate(args):
                if os.path.isfile(arg):
                    files.append(arg)
                    args[idx] = arg.split("/")[-1]

        files_data = []
        for f in files:
            path = f
            if not f.startswith("/"):
                # make path absolute
                path = cwd + "/" + f

            arg_split = path.split("/")
            file = open(path, "rb")
            files_data.append({
                "pathWithoutFilename": "/",
                "name": arg_split[-1],
                "data": python_bytes_to_byte64(file.read())
            })
            file.close()

        stdin = python_bytes_to_byte64(stdin)

        BioLibServer.log_level = BioLib.log_level
        with BioLibServer(("127.0.0.1", 0), BioLibHandler) as httpd:
            port = httpd.server_address[1]
            thread = threading.Thread(target=httpd.serve_forever)
            # TODO: figure out how we can avoid changing the current directory
            os.chdir(os.path.dirname(os.path.realpath(__file__)) + "/biolib-js/")
            try:
                thread.start()
                res = asyncio.get_event_loop().run_until_complete(
                    self.call_pyppeteer(port=port, args=args, stdin=stdin, files=files_data, output_path=output_path))
            finally:
                os.chdir(cwd)
            httpd.shutdown()
            thread.join()
            return res

        raise BioLibError("Failed to start TCPServer")



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

"""Coverage module."""
import asyncio
import logging
from functools import cmp_to_key
from typing import Dict, List

from biolib.pyppeteer.pyppeteer import helpers
from biolib.pyppeteer.pyppeteer.connection import CDPSession
from biolib.pyppeteer.pyppeteer.errors import PageError
from biolib.pyppeteer.pyppeteer.execution_context import EVALUATION_SCRIPT_URL
from biolib.pyppeteer.pyppeteer.models import CoverageResult, NestedRangeItemInput, NestedRangeItem, Protocol


logger = logging.getLogger(__name__)


class Coverage:
    """Coverage class.

    Coverage gathers information about parts of JavaScript and CSS that were
    used by the page.

    An example of using JavaScript and CSS coverage to get percentage of
    initially executed code::

        # Enable both JavaScript and CSS coverage
        await page.coverage.startJSCoverage()
        await page.coverage.startCSSCoverage()

        # Navigate to page
        await page.goto('https://example.com')
        # Disable JS and CSS coverage and get results
        jsCoverage = await page.coverage.stopJSCoverage()
        cssCoverage = await page.coverage.stopCSSCoverage()
        totalBytes = 0
        usedBytes = 0
        coverage = jsCoverage + cssCoverage
        for entry in coverage:
            totalBytes += len(entry['text'])
            for range in entry['ranges']:
                usedBytes += range['end'] - range['start'] - 1

        print('Bytes used: {}%'.format(usedBytes / totalBytes * 100))
    """

    def __init__(self, client: CDPSession) -> None:
        self._jsCoverage = JSCoverage(client)
        self._cssCoverage = CSSCoverage(client)

    async def startJSCoverage(self, resetOnNavigation: bool = True, reportAnonymousScripts: bool = False,) -> None:
        """Start JS coverage measurement.

        :param resetOnNavigation: Whether to reset coverage on every
          navigation.
        :param reportAnonymousScripts: Whether anonymous script generated
          by the page should be reported.

        .. note::
            Anonymous scripts are ones that don't have an associated url. These
            are scripts that are dynamically created on the page using ``eval``
            of ``new Function``. If ``reportAnonymousScript`` is set to
            ``True``, anonymous scripts will have
            ``__pyppeteer_evaluation_script__`` as their url.
        """
        await self._jsCoverage.start(
            resetOnNavigation=resetOnNavigation, reportAnonymousScripts=reportAnonymousScripts,
        )

    async def stopJSCoverage(self) -> List:
        """Stop JS coverage measurement and get result.

        Return list of coverage reports for all scripts. Each report includes:

        * ``url`` (str): Script url.
        * ``text`` (str): Script content.
        * ``ranges`` (List[Dict]): Script ranges that were executed. Ranges are
          sorted and non-overlapping.

          * ``start`` (int): A start offset in text, inclusive.
          * ``end`` (int): An end offset in text, exclusive.

        .. note::
           JavaScript coverage doesn't include anonymous scripts by default.
           However, scripts with sourceURLs are reported.
        """
        return await self._jsCoverage.stop()

    async def startCSSCoverage(self, resetOnNavigation: bool = True) -> None:
        """Start CSS coverage measurement.


        :param resetOnNavigation: Whether to reset coverage on every
          navigation.
        """
        await self._cssCoverage.start(resetOnNavigation=resetOnNavigation)

    async def stopCSSCoverage(self) -> List:
        """Stop CSS coverage measurement and get result.

        Return list of coverage reports for all non-anonymous scripts. Each
        report includes:

        * ``url`` (str): StyleSheet url.
        * ``text`` (str): StyleSheet content.
        * ``ranges`` (List[Dict]): StyleSheet ranges that were executed. Ranges
          are sorted and non-overlapping.

          * ``start`` (int): A start offset in text, inclusive.
          * ``end`` (int): An end offset in text, exclusive.

        .. note::
           CSS coverage doesn't include dynamically injected style tags without
           sourceURLs (but currently includes... to be fixed).
        """
        return await self._cssCoverage.stop()


class JSCoverage:
    """JavaScript Coverage class."""

    def __init__(self, client: CDPSession) -> None:
        self._client = client
        self._enabled = False
        self._scriptURLs = {}
        self._scriptSources = {}
        self._eventListeners = []
        self._resetOnNavigation = False

    async def start(self, resetOnNavigation: bool = True, reportAnonymousScripts: bool = False,) -> None:
        """Start coverage measurement."""
        if self._enabled:
            raise PageError('JSCoverage is already enabled.')
        self._resetOnNavigation = resetOnNavigation
        self._reportAnonymousScript = reportAnonymousScripts
        self._enabled = True
        self._scriptURLs.clear()
        self._scriptSources.clear()
        self._eventListeners = [
            helpers.addEventListener(
                self._client, 'Debugger.scriptParsed', lambda e: self._onScriptParsed(e)
            ),
            helpers.addEventListener(
                self._client, 'Runtime.executionContextsCleared', self._onExecutionContextsCleared
            ),
        ]
        await asyncio.gather(
            self._client.send('Profiler.enable'),
            self._client.send('Profiler.startPreciseCoverage', {'callCount': False, 'detailed': True}),
            self._client.send('Debugger.enable'),
            self._client.send('Debugger.setSkipAllPauses', {'skip': True}),
        )

    def _onExecutionContextsCleared(self, event: Dict) -> None:
        if not self._resetOnNavigation:
            return
        self._scriptURLs.clear()
        self._scriptSources.clear()

    async def _onScriptParsed(self, event: Protocol.Debugger.scriptParsedPayload) -> None:
        # Ignore pyppeteer-injected scripts
        if event.get('url') == EVALUATION_SCRIPT_URL:
            return
        # Ignore other anonymous scripts unless the reportAnonymousScript
        # option is True
        if not event.get('url') and not self._reportAnonymousScript:
            return

        scriptId = event.get('scriptId')
        url = event.get('url')
        try:
            response = await self._client.send('Debugger.getScriptSource', {'scriptId': scriptId})
            self._scriptURLs[scriptId] = url
            self._scriptSources[scriptId] = response.get('scriptSource')
        except Exception as e:
            logger.error(f'An exception occurred during _onScriptParsed handling: {e}'
                         f'\nThis might happen if the page has already navigated away.')

    async def stop(self) -> List:
        """Stop coverage measurement and return results."""
        if not self._enabled:
            raise PageError('JSCoverage is not enabled')
        self._enabled = False

        result, *_ = await asyncio.gather(
            self._client.send('Profiler.takePreciseCoverage'),
            self._client.send('Profiler.stopPreciseCoverage'),
            self._client.send('Profiler.disable'),
            self._client.send('Debugger.disable'),
        )
        helpers.removeEventListeners(self._eventListeners)

        coverage = []
        for entry in result.get('result', []):
            scriptId = entry.get('scriptId')
            url = self._scriptURLs.get(scriptId)
            if not url and self._reportAnonymousScript:
                url = f'debugger://VM{scriptId}'
            text = self._scriptSources.get(scriptId)
            if text is None or url is None:
                continue
            flattenRanges = []
            for func in entry.get('functions', []):
                flattenRanges.extend(func.get('ranges', []))
            ranges = convertToDisjointRanges(flattenRanges)
            coverage.append({'url': url, 'ranges': ranges, 'text': text})
        return coverage


class CSSCoverage:
    """CSS Coverage class."""

    def __init__(self, client: CDPSession) -> None:
        self._client = client
        self._enabled = False
        self._stylesheetURLs: Dict = {}
        self._stylesheetSources: Dict = {}
        self._eventListeners: List = []
        self._resetOnNavigation = False

    async def start(self, resetOnNavigation: bool = True) -> None:
        """Start coverage measurement."""
        if self._enabled:
            raise PageError('CSSCoverage is already enabled.')
        self._resetOnNavigation = resetOnNavigation
        self._enabled = True
        self._stylesheetURLs.clear()
        self._stylesheetSources.clear()
        self._eventListeners = [
            helpers.addEventListener(
                self._client, 'CSS.styleSheetAdded', lambda e: self._client.loop.create_task(self._onStyleSheet(e))
            ),
            helpers.addEventListener(
                self._client, 'Runtime.executionContextsCleared', self._onExecutionContextsCleared
            ),
        ]
        await asyncio.gather(
            self._client.send('DOM.enable'),
            self._client.send('CSS.enable'),
            self._client.send('CSS.startRuleUsageTracking'),
        )

    def _onExecutionContextsCleared(self, event: Dict) -> None:
        if not self._resetOnNavigation:
            return
        self._stylesheetURLs.clear()
        self._stylesheetSources.clear()

    async def _onStyleSheet(self, event: Dict) -> None:
        header = event.get('header', {})
        # Ignore anonymous scripts
        if not header.get('sourceURL'):
            return
        try:
            response = await self._client.send('CSS.getStyleSheetText', {'styleSheetId': header['styleSheetId']})
            self._stylesheetURLs[header['styleSheetId']] = header['sourceURL']
            self._stylesheetSources[header['styleSheetId']] = response['text']
        except Exception as e:
            # This might happen if the page has already navigated away.
            logger.error(f'An exception occurred: {e}')

    async def stop(self) -> List[CoverageResult]:
        """Stop coverage measurement and return results."""
        if not self._enabled:
            raise PageError('CSSCoverage is not enabled.')
        self._enabled = False
        ruleTrackingResponse = await self._client.send('CSS.stopRuleUsageTracking')
        await asyncio.gather(self._client.send('CSS.disable'), self._client.send('DOM.disable'))
        helpers.removeEventListeners(self._eventListeners)

        # aggregate by styleSheetId
        styleSheetIdToCoverage: Dict[str, List[Dict[str, str]]] = {}
        for entry in ruleTrackingResponse['ruleUsage']:
            ranges = styleSheetIdToCoverage.get(entry['styleSheetId'])
            if not ranges:
                ranges = []
                styleSheetIdToCoverage[entry['styleSheetId']] = ranges
            ranges.append(
                {
                    'startOffset': entry['startOffset'],
                    'endOffset': entry['endOffset'],
                    'count': 1 if entry['used'] else 0,
                }
            )

        coverage = []
        for styleSheetId in self._stylesheetURLs:
            url = self._stylesheetURLs.get(styleSheetId)
            text = self._stylesheetSources.get(styleSheetId)
            ranges = convertToDisjointRanges(styleSheetIdToCoverage.get(styleSheetId, []))
            coverage.append({'url': url, 'ranges': ranges, 'text': text})

        return coverage


def convertToDisjointRanges(nestedRanges: List[NestedRangeItemInput]) -> List[NestedRangeItem]:
    """
    Convert ranges.
    NestedRange members support keys:
        * startOffset: float
        * endOffset: float
        * count: int
    """
    points: List = []
    for nested_range in nestedRanges:
        points.append({'offset': nested_range['startOffset'], 'type': 0, 'range': nested_range})
        points.append({'offset': nested_range['endOffset'], 'type': 1, 'range': nested_range})

    # Sort points to form a valid parenthesis sequence.
    def _sort_func(a: Dict, b: Dict) -> int:
        # Sort with increasing offsets.
        if a['offset'] != b['offset']:
            return a['offset'] - b['offset']
        # All "end" points should go before "start" points.
        if a['type'] != b['type']:
            return b['type'] - a['type']
        aLength = a['range']['endOffset'] - a['range']['startOffset']
        bLength = b['range']['endOffset'] - b['range']['startOffset']
        # For two "start" points, the one with longer range goes first.
        if a['type'] == 0:
            return bLength - aLength
        # For two "end" points, the one with shorter range goes first.
        return aLength - bLength

    points.sort(key=cmp_to_key(_sort_func))

    hitCountStack = []
    results = []
    lastOffset = 0
    # Run scanning line to intersect all ranges.
    for point in points:
        if hitCountStack and lastOffset < point['offset'] and hitCountStack[len(hitCountStack) - 1] > 0:
            lastResult = results[-1] if results else None
            if lastResult and lastResult['end'] == lastOffset:
                lastResult['end'] = point['offset']
            else:
                results.append({'start': lastOffset, 'end': point['offset']})
        lastOffset = point['offset']
        if point['type'] == 0:
            hitCountStack.append(point['range']['count'])
        else:
            hitCountStack.pop()
    # Filter out empty ranges.
    return [range for range in results if range['end'] - range['start'] > 1]



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

import subprocess
from os import path

from livereload import Server, watcher

watcher.pyinotify = None  # disable pyinotify

docsdir = path.dirname(path.abspath(__file__))
builddir = path.join(docsdir, '_build')
build_cmd = [
    'sphinx-build',
    '-q',
    '-j',
    'auto',
    '-b',
    'html',
    '-d',
    path.join(builddir, 'doctrees'),
    docsdir,
    path.join(builddir, 'html'),
]


def cmd() -> None:
    print('=== Sphinx Build Start ===')
    subprocess.run(build_cmd, cwd=docsdir)
    print('=== Sphinx Build done ===')


# subprocess.run(['make', 'clean'], cwd=docsdir)
cmd()
server = Server()


def docs(p: str) -> str:
    return path.join(docsdir, p)


# Watch documents
server.watch(docs('*.py'), cmd, delay=1)
server.watch(docs('*.md'), cmd, delay=1)
server.watch(docs('../*.md'), cmd, delay=1)
server.watch(docs('*.md'), cmd, delay=1)
server.watch(docs('*/*.md'), cmd, delay=1)
server.watch(docs('*/*/*.md'), cmd, delay=1)

# Watch template/style
server.watch(docs('_templates/*.html'), cmd, delay=1)
server.watch(docs('_static/*.css'), cmd, delay=1)
server.watch(docs('_static/*.js'), cmd, delay=1)

# Watch package
server.watch(docs('../pyppeteer/*.py'), cmd, delay=1)
server.watch(docs('../pyppeteer/*/*.py'), cmd, delay=1)
server.watch(docs('../pyppeteer/*/*/*.py'), cmd, delay=1)

server.serve(port=8889, root=docs('_build/html'), debug=True, restart_delay=1)
